/********************
 * 467 GPU Assembler
 *
 *
 ********************/

#include <stdbool.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "467gpu.h"

//////////////////////////////////////////////////
///
///
/// It is less likely anything below here will have to
/// be edited much.
///
static bool _iswhitespace(char ch) {
    switch (ch) {
        case    '/':
        case    ',':
        case    ' ':
        case    '\t':
            return true;
        default:
            return false;
    }
}

static bool _iscommentstart(char ch) {
    switch (ch) {
        case    '#':
            return true;
        default:
            return false;
    }
}

static bool _ishardnewline(char ch) {
    switch (ch) {
        case    '\0':
        case    '\n':
            return true;
        default:
            return false;
    }
}

static bool _isnewline(char ch) {
    switch (ch) {
        case    '\0':
        case    '\n':
        case    ';':
            return true;
        default:
            return false;
    }
}

static char _lowercase(char ch) {
    if (ch >= 'A' && ch <= 'Z')
        return ch - 'A' + 'a';
    return ch;
}

static int line_number = 1;

// Read a line of input.  But strip out useless whitespace and comments
// and also break multi-line statements into separate lines.  Also
// convert all characters to lower case.
static int _readline(char *s, int maxlen, FILE *infp) {
    int count = 0;
    char    ch;
    
    while (count < maxlen) {
        if (feof(infp)) {
            if (count > 0)
                return count;
            else
                return -1;
        }
        ch = fgetc(infp);
        if (feof(infp) && count == 0)
            return -1;
            
        // Update the line number if needed
        if (_ishardnewline(ch))
            ++line_number;
            
        // Skip white space at the start of a line
        if ((_iswhitespace(ch) || _isnewline(ch)) && count == 0)
            continue;
        
        // if this is a comment, skip the input until the of the line
        if (_iscommentstart(ch) && count == 0) {
            fgetc(infp);  // discard character
            fgets(s, maxlen, infp);
            int temp_line_number = atoi(s);
            if (temp_line_number > 0)
                line_number = temp_line_number;
            continue;
        }
        
        // If we are starting a comment or ending a line and we have
        // characters, then return.
        if ((_isnewline(ch) || _iscommentstart(ch)) && count != 0) {
            *s = '\0';
            return count;
        }
        
        *s = ch;
        ++s;
        ++count;
    }
    
    printf("Long line on input\n");
    exit(-__LINE__);
}

static void _parse_operand(u_int32_t *constant,
    u_int32_t   *reg1,
    u_int32_t   *reg2,
    u_int32_t   *reg3,
    bool        *use_constant,
    char        *s,
    u_int32_t   operand_format,
    char        *input_line) {

    // perhaps there really shouldn't be an operand here
    if (operand_format == 0) {
        if (strlen(s) != 0) {
            printf("Malformed operand on line: (%d) %s\n", line_number, input_line);
            exit(-__LINE__);
        }
        return;
    }

    // we sloppily allow predicates to be specified as r or p
    operand_format = operand_format & (~PREDICATE);
    
    // depending on whether this is a memory or a register operand
    // discard the proper prefix.  * for memory and r or $ for registers
    if ((operand_format & MEMORY) == MEMORY) {
        operand_format = operand_format & (~MEMORY);
        if (*s != '*') {
            printf("Malformed operand on line: (%d) %s\n", line_number, input_line);
            exit(-__LINE__);
        }
        ++s;
        if (operand_format != CONSTANT) {
          ++s;
        }
    } else {
        if (operand_format != CONSTANT) {
            if (!(*s == 'r' || *s == 'R' || *s == '$' || *s == 'p' || *s == 'P')) {
                printf("Malformed operand on line: (%d) %s\n", line_number, input_line);
                exit(-__LINE__);
            }
            ++s;
        }
    }
    
    if (operand_format == CONSTANT) {
        u_int32_t   v = strtoul(s, NULL, 0);
        v = v & MAX_CONSTANT;
        *constant = v;
        *use_constant = true;
        return;    
    }
    
    if (operand_format == REG1)
        *reg1 = atoi(s);
    else if (operand_format == REG2)
        *reg2 = atoi(s);
    else if (operand_format == REG3)
        *reg3 = atoi(s);
    else {
        printf("Malformed assembly format table for instruction: %s\n", input_line);
        exit(-__LINE__);
    }
}

enum {
    output_format_ascii_hex = 1,
    output_format_ascii_binary,
    output_format_binary,
};

int main(int argc, char *argv[]) {
    FILE    *infp = stdin;
    FILE    *outfp = stdout;
    FILE    *debugfp = NULL;
    int     output_format = output_format_ascii_hex;
    int     i;
    
    i = 1;
    while (i < argc) {
        if (argv[i][0] == '-') {
            switch(argv[i][1]) {
                case    'i':
                case    'I':
                    infp = fopen(argv[i + 1], "r");
                    if (!infp) {
                        printf("Failure to open input file: %s\n", argv[i + 1]);
                        exit(__LINE__);
                    }
                    i += 2;
                    break;
                case    'o':
                case    'O':
                    outfp = fopen(argv[i + 1], "w");
                    if (!outfp) {
                        printf("Failure to create output file: %s\n", argv[i + 1]);
                        exit(__LINE__);
                    }
                    i += 2;
                    break;
                case    'd':
                case    'D':
                    debugfp = fopen(argv[i + 1], "w");
                    if (!debugfp) {
                        printf("Failure to create debug symbols file: %s\n", argv[i + 1]);
                        exit(__LINE__);
                    }
                    i += 2;
                    break;
                case    'a':
                case    'A':
                    output_format = output_format_ascii_hex;
                    ++i;
                    break;
                case    'b':
                case    'B':
                    output_format = output_format_ascii_binary;
                    ++i;
                    break;
                case    'm':
                case    'M':
                    output_format = output_format_binary;
                    ++i;
                    break;
                default:
                    printf("Usage:\n"
                        "\t-i <input>     input file\n"
                        "\t-o <output>    output file\n"
                        "\t-d <file>      debug symbols file\n"
                        "\t-a             output ascii hex\n"
                        "\t-b             output ascii binary\n"
                        "\t-m             output binary\n");
                    exit(__LINE__);
            }
        } else {
            printf("Try -- for help\n");
            exit(-__LINE__);
        }
    }

    char    input_line[MAX_LINE_LENGTH];
    
#define MAX_TOKENS  5
    char    tokens[MAX_TOKENS][MAX_LINE_LENGTH];
    int token;
    char    *s;
    
    while (_readline(input_line, MAX_LINE_LENGTH, infp) > 0) {
        // Parse the input line into individual tokens
        i = 0;
        for (token = 0; token < MAX_TOKENS; token++)
            tokens[token][0] = '\0';
        token = 0;
        s = input_line;
        while (*s) {
            if (_iswhitespace(*s)) {
                tokens[token][i] = '\0';
                i = 0;
                ++token;
                while (*s && _iswhitespace(*s))
                    ++s;
                if (!(*s))
                    break;
                if (token >= MAX_TOKENS) {
                    if (!(*s)) {
                        printf("Cannot parse: (%d) %s\n", line_number, input_line);
                        exit(-__LINE__);
                    }
                    break;
                }
            }
            tokens[token][i] = *s;
            ++i;
            ++s;
        }
        if (i > 0)
            tokens[token][i] = '\0';

        // Match the mnemonic (token0) to the parse table
        
        int ins = 0;
        while (instructions[ins].mnemonic) {
            if (!strcmp(instructions[ins].mnemonic, tokens[0]))
                break;
            ++ins;
        }
        if (!instructions[ins].mnemonic) {
            printf("Cannot parse: (%d) %s\n", line_number, input_line);
            exit(-__LINE__);
        }
        
        // The table now tells us how to parse the remaining three tokens
        
        u_int32_t   opcode = instructions[ins].opcode;
        u_int32_t   constant = 0;
        u_int32_t   reg1 = 0, reg2 = 0, reg3 = 0;
        u_int32_t   predicate = 0;
        bool        use_constant = false;
        int         with_predicate = 0;
        u_int32_t   mask;
        
        if (strchr(input_line, '/')) {        
            predicate = atoi(tokens[1]);
            with_predicate = 1;
        }
            
        _parse_operand(&constant, &reg1, &reg2, &reg3, &use_constant,
            tokens[1 + with_predicate], instructions[ins].operand0, input_line);
        _parse_operand(&constant, &reg1, &reg2, &reg3, &use_constant,
            tokens[2 + with_predicate], instructions[ins].operand1, input_line);
        _parse_operand(&constant, &reg1, &reg2, &reg3, &use_constant,
            tokens[3 + with_predicate], instructions[ins].operand2, input_line);
            
        u_int32_t   res = encode_instruction(
            opcode, predicate, use_constant, constant,
            reg1, reg2, reg3);

        switch (output_format) {
            case    output_format_ascii_hex:
                fprintf(outfp, "0x%8.8x\n", res);
                break;
            case    output_format_ascii_binary:
                i = 0;
                
                mask = (1 << 31);
                while (i < 32) {
                    if (res & mask)
                        fprintf(outfp, "1");
                    else
                        fprintf(outfp, "0");
                    mask = mask >> 1;
                    ++i;
                }
                fprintf(outfp, "\n");
                break;
            case    output_format_binary:
                fwrite(&res, 1, sizeof(res), outfp);
                break;
        }
        
        if (debugfp) {
            struct  debug_symbol   ds;
            
            memset(&ds, 0, sizeof(ds));
            ds.line_number = line_number;
            ds.line_length = strlen(input_line) + 1;
            strcpy(ds.line, input_line);
            fwrite(&ds, 1, sizeof(int) * 2 + strlen(input_line) + 1, debugfp);
        }
    }

    if (infp != stdin)
        fclose(infp);
    if (outfp != stdout)
        fclose(outfp);
        
    return 0;    
}
