#include <stdio.h>
#include <stdlib.h>
#include <stdbool.h>
#include <string.h>
#include "467gpu.h"

// Debugger
static bool _iswhitespace(char ch) {
    switch (ch) {
        case    '/':
        case    ',':
        case    ' ':
        case    '\t':
        case    '\n':
            return true;
        default:
            return false;
    }
}

static int _tokenize(char *s, char **tokens, int max_tokens) {
    // Parse the input line into individual tokens
    int i = 0;
    int token;
    for (token = 0; token < max_tokens; token++)
        tokens[token][0] = '\0';
    token = 0;

    while (*s) {
        if (_iswhitespace(*s)) {
            tokens[token][i] = '\0';
            i = 0;
            ++token;
            while (*s && _iswhitespace(*s))
                ++s;
            if (!(*s))
                break;
            if (token >= max_tokens)
                return token;
        }
        tokens[token][i] = *s;
        ++i;
        ++s;
    }

    if (i > 0)
        tokens[token][i] = '\0';
    ++token;
    return token;
}

enum breakpoint_t {
    disabled    =   0,
    pc_equal,
    reg_value_equal,
    reg_value_lt,
    reg_value_le,
    reg_value_gt,
    reg_value_ge,
    mem_value_equal,
    mem_value_lt,
    mem_value_le,
    mem_value_gt,
    mem_value_ge
};

#define NUMBER_OF_BREAKPOINTS   8

struct breakpoint {
    enum breakpoint_t   breakpoint_type;
    u_int32_t   value;
    u_int32_t   address;
};

struct  execution_context {
    struct  debug_symbol    source_file[CODE_SIZE];
    struct  debug_symbol    symbol_file[CODE_SIZE];
    u_int32_t   code[CODE_SIZE];
    u_int32_t   data[GPMEM_SIZE];
    struct      context ctx;
    int         code_size;
    struct  breakpoint  breakpoints[NUMBER_OF_BREAKPOINTS];
};

static bool _is_at_specific_breakpoint(struct execution_context *ec,
    struct breakpoint *bp) {
    switch (bp->breakpoint_type) {
        case    disabled:   return false;
        case    pc_equal:   return ec->ctx.pc == bp->value;
        case    reg_value_equal:    return ec->ctx.regs[bp->address] == bp->value;
        case    reg_value_lt:       return ec->ctx.regs[bp->address] < bp->value;
        case    reg_value_le:       return ec->ctx.regs[bp->address] <= bp->value;
        case    reg_value_gt:       return ec->ctx.regs[bp->address] > bp->value;
        case    reg_value_ge:       return ec->ctx.regs[bp->address] >= bp->value;
        case    mem_value_equal:    return ec->data[bp->address] == bp->value;
        case    mem_value_lt:       return ec->data[bp->address] < bp->value;
        case    mem_value_le:       return ec->data[bp->address] <= bp->value;
        case    mem_value_gt:       return ec->data[bp->address] > bp->value;
        case    mem_value_ge:       return ec->data[bp->address] >= bp->value;
        default: return false;
    }
}

static void _print_breakpoint(struct breakpoint *bp, FILE *outfp) {
    switch (bp->breakpoint_type) {
        case    disabled:
            break;
        case    pc_equal:
            printf("pc = 0x%x", bp->value);
            break;
        case    reg_value_equal:
            printf("r%d = 0x%x", bp->address, bp->value);
            break;
        case    reg_value_lt:
            printf("r%d < 0x%x", bp->address, bp->value);
            break;
        case    reg_value_le:
            printf("r%d <= 0x%x", bp->address, bp->value);
            break;
        case    reg_value_gt:
            printf("r%d > 0x%x", bp->address, bp->value);
            break;
        case    reg_value_ge:
            printf("r%d >= 0x%x", bp->address, bp->value);
            break;
        case    mem_value_equal:
            printf("mem[0x%x] = 0x%x", bp->address, bp->value);
            break;
        case    mem_value_lt:
            printf("mem[0x%x] < 0x%x", bp->address, bp->value);
            break;
        case    mem_value_le:
            printf("mem[0x%x] <= 0x%x", bp->address, bp->value);
            break;
        case    mem_value_gt:
            printf("mem[0x%x] > 0x%x", bp->address, bp->value);
            break;
        case    mem_value_ge:
            printf("mem[0x%x] >= 0x%x", bp->address, bp->value);
            break;
    }
}

static void _print_reached_breakpoint(struct execution_context *ec, FILE *outfp) {
    int i;

    fprintf(outfp, "The following breakpoint(s) have been reached:\n\n");

    for (i = 0; i < NUMBER_OF_BREAKPOINTS; i++)
        if (_is_at_specific_breakpoint(ec, &ec->breakpoints[i])) {
            fprintf(outfp, "breakpoint %d: ", i);
            _print_breakpoint(&ec->breakpoints[i], outfp);
            fprintf(outfp, "\n");
            }
}

static bool _is_at_any_breakpoint(struct execution_context *ec) {
    int i;
    
    for (i = 0; i < NUMBER_OF_BREAKPOINTS; i++)
        if (_is_at_specific_breakpoint(ec, &ec->breakpoints[i]))
            return true;
    return false;
}

static void _read_source_file(struct execution_context *ec, FILE *fp) {
    int i;
    
    i = 0;
    while (!feof(fp)) {
        fgets(ec->source_file[i].line, MAX_LINE_LENGTH, fp);
        ec->source_file[i].line_number = i + 1;
        ec->source_file[i].line_length = strlen(ec->source_file[i].line);
        ++i;
    }
}

static void _read_debug_symbols(struct execution_context *ec, FILE *fp) {
    int i;
    
    i = 0;
    while (!feof(fp)) {
        if (fread(&ec->symbol_file[i], 2, sizeof(int), fp) != (2 * sizeof(int)))
            break;
        if (ec->symbol_file[i].line_length > MAX_LINE_LENGTH) {
            printf("Error reading symbol file\n");
            exit(-__LINE__);
        }
        fread(&ec->symbol_file[i].line, ec->symbol_file[i].line_length, 1, fp);
        ++i;
    }
}

static void _read_code(struct execution_context *ec, FILE *fp) {
    ec->code_size = fread(ec->code, CODE_SIZE, sizeof(u_int32_t), fp) / sizeof(u_int32_t);
}

static void _read_memory_binary(struct execution_context *ec, FILE *fp) {
    fread(ec->data, GPMEM_SIZE, sizeof(u_int32_t), fp);
}

static bool _isdigit(char ch) {
    if (ch >= '0' && ch <= '9')
        return true;
    return false;
}

static bool _ishexdigit(char ch) {
    if ((ch >= '0' && ch <= '9') || (ch >= 'A' && ch <= 'F') || (ch >= 'a' && ch <= 'f'))
        return true;
    return false;
}

static void _read_memory_ascii(struct execution_context *ec, FILE *fp) {
    u_int32_t   i;
    u_int32_t   v;
    char    s[MAX_LINE_LENGTH];
    char    *str;
    // expect 1 integer per line.  But also allow for an address set with:
    //
    // address:    // no other things permitted
    
    while (!feof(fp)) {
        fgets(s, MAX_LINE_LENGTH, fp);
        if (feof(fp))
            return;
        if ((str = strchr(s, ':'))) {
            // this is not a value, this is an address change.
            i = strtoul(s, NULL, 16);
            str += 2;
        } else {
            str = s;
        }
        if (!_ishexdigit(str[0]))
            continue;
        if (i >= GPMEM_SIZE) {
            printf("Memory image file corrupt address: 0x%x\n", i);
            exit(-__LINE__); 
        }
        v = strtoul(str, NULL, 16);
        ec->data[i] = v;
        ++i;
    }
    
}

void usage_and_exit() {
    printf("usage:\n"
        "\t-b <file>   binary image file\n"
        "\t-d <file>   debug symbol file\n"
        "\t-s <file>   source file\n"
        "\t-m <file>   memory image file (binary)\n"
        "\t-a <file>   memory image file (ascii)\n"
        "\t-i <file>   input command file (default = stdin)\n"
        "\t-c          after executing input file, continue with stdin for input\n"
        "\t-o <file>   send output to this file\n"
        "\t-h          this message\n");
    exit(-__LINE__);
}

struct  debug_command {
    char    *command;
    char    *abbreviation;
    void    (*func)(FILE *outfp, struct execution_context *, char **tokens);
    char    *help;
};

// not really extern, just a forward declaration
extern struct debug_command debug_commands[];

void _print_debug_commands(FILE *outfp, struct execution_context *ec, char **tokens) {
    int i;
    
    i = 0;
    while (debug_commands[i].command) {
        fprintf(outfp, "%s(%s)\t%s\n",
            debug_commands[i].command, debug_commands[i].abbreviation,
            debug_commands[i].help);
        ++i;
    }
}

void _print_registers(FILE *outfp, struct execution_context *ec, char **tokens) {
    int i = 0;
    while (i < NUMBER_OF_REGISTERS) {
        if (i < 12)
            fprintf(outfp, "%x: %8.8x    %x: %8.8x    %x: %8.8x    %x: %8.8x\n",
                i, ec->ctx.regs[i],
                i+1, ec->ctx.regs[i+1],
                i+2, ec->ctx.regs[i+2],
                i+3, ec->ctx.regs[i+3]);
        else
            fprintf(outfp, "%x: %8.8x    %x: %8.8x    %x: %8.8x\n",
                i, ec->ctx.regs[i],
                i+1, ec->ctx.regs[i+1],
                i+2, ec->ctx.regs[i+2]);
                
        i += 4;
    }
    fprintf(outfp, "p0: %d p1: %d p2: %d p3: %d src: %d pid: 0x%2.2x pc: 0x%4.4x\n",
        ec->ctx.predicate0,
        ec->ctx.predicate1,
        ec->ctx.predicate2,
        ec->ctx.predicate3,
        ec->ctx.queue_src,
        ec->ctx.pid,
        ec->ctx.pc);
}

void _quit(FILE *outfp, struct execution_context *ec, char **tokens) {
    exit(0);
}

void _set_register(FILE *outfp, struct execution_context *ec, char **tokens) {
    u_int32_t   r, v;
    
    v = strtoul(tokens[2], NULL, 0);
    if (!strcmp(tokens[1], "p0"))
        ec->ctx.predicate0 = v;
    else if (!strcmp(tokens[1], "p1"))
        ec->ctx.predicate1 = v;
    else if (!strcmp(tokens[1], "p2"))
        ec->ctx.predicate2 = v;
    else if (!strcmp(tokens[1], "p1"))
        ec->ctx.predicate1 = v;
    else if (!strcmp(tokens[1], "pid"))
        ec->ctx.pid = v;
    else if (!strcmp(tokens[1], "src"))
        ec->ctx.queue_src = v;
    else if (!strcmp(tokens[1], "pc"))
        ec->ctx.pc= v;
    else if (tokens[1][0] == 'r' || tokens[1][0] == 'R') {
        r = strtoul(&tokens[1][1], NULL, 0);
        if (r >= (NUMBER_OF_REGISTERS-1))
            fprintf(outfp, "Invalid register to set: %d", r);
        else
            ec->ctx.regs[r] = v;
        }
    else
        fprintf(outfp, "Unknown register: %s\n", tokens[1]);
}

void _single_step(FILE *outfp, struct execution_context *ec, char **tokens) {
    enum single_step_result_t r;
    u_int32_t   output[QUEUE_IO];
    int         output_queue;
    int         i;
    char        s[MAX_LINE_LENGTH];
    
    if (!disasm_instruction(ec->code[ec->ctx.pc], s)) {
        fprintf(outfp, "Failure to disassemble instruction.");
        return;
    }
    fprintf(outfp, "0x%x:    %s\n", ec->ctx.pc, s); 
    r = single_step(&ec->ctx, ec->code, ec->data, output, &output_queue);
    switch (r) {
        case    at_end:
            fprintf(outfp, "END reached\n");
            break;
        case    normal_execution:
            break;
        case    skipped_execution:
            fprintf(outfp, "predicated to no-op\n");
            break;
        case    queue_write:
            fprintf(outfp, "Queue write: %d\n", output_queue);
            for (i = 0; i < QUEUE_IO; i++)
                fprintf(outfp, "\t%d: %8.8x\n", i, output[i]);
            break;
        case    invalid_predicate:
            fprintf(outfp, "ERROR: invalid predicate on execution");
            break;
        case    invalid_address:
            fprintf(outfp, "ERROR: attempt to read/write invalid memory address\n");
            break;
    }
    if (_is_at_any_breakpoint(ec))
        _print_reached_breakpoint(ec, outfp);
}

void _run(FILE *outfp, struct execution_context *ec, char **tokens) {
    enum single_step_result_t r;
    u_int32_t   output[QUEUE_IO];
    int         output_queue;
    int         i;

    while (true) {    
        r = single_step(&ec->ctx, ec->code, ec->data, output, &output_queue);
        switch (r) {
            case    at_end:
                fprintf(outfp, "END reached\n");
                return;
            case    normal_execution:
                break;
            case    skipped_execution:
                break;
            case    queue_write:
                fprintf(outfp, "Queue write: %d\n", output_queue);
                for (i = 0; i < QUEUE_IO; i++)
                    fprintf(outfp, "\t%d: %8.8x\n", i, output[i]);
                break;
            case    invalid_predicate:
                fprintf(outfp, "ERROR: invalid predicate on execution");
                return;
                break;
            case    invalid_address:
                fprintf(outfp, "ERROR: attempt to read/write invalid memory address\n");
                return;
                break;
        }
        if (_is_at_any_breakpoint(ec)) {
            _print_reached_breakpoint(ec, outfp);
            break;
        }
    }
}


void _disasm(FILE *outfp, struct execution_context *ec, char **tokens) {
    char        s[MAX_LINE_LENGTH];
    int32_t   addr;
    int32_t   len = 16;
    int32_t   i;
    
    if (strlen(tokens[1]) >= 1)
        addr = strtoul(tokens[1], NULL, 0);
    else {
        addr = ec->ctx.pc - len/2;
        if (addr < 0)
            addr = 0;
    }
    i = 0;
    while (i < len) {
        if (addr >= CODE_SIZE)
            return;
        if (!disasm_instruction(ec->code[addr], s)) {
            return;
        }
        if (addr == ec->ctx.pc)
            fprintf(outfp, "* ");
        fprintf(outfp, "0x%x:    %s\n", addr, s);
        ++i;
        ++addr;
    }
}

void _dump_memory(FILE *outfp, struct execution_context *ec, char **tokens) {
    if (strlen(tokens[2]) < 1) {
        printf("usage: dump address [length]\n");
        return;
    }
    u_int32_t   addr = strtoul(tokens[1], NULL, 0);
    u_int32_t   len = 16, i;
    if (strlen(tokens[2]) > 0)
        len = strtoul(tokens[2], NULL, 0);
    i = 0;
    while (i < len) {
        if (addr >= GPMEM_SIZE)
            return;
        fprintf(outfp, "0x%x: %8.8x\n", addr, ec->data[addr]);
        ++addr;
        ++i;
    }
}

void _set_memory(FILE *outfp, struct execution_context *ec, char **tokens) {
    if (strlen(tokens[2]) < 1) {
        printf("usage: setm address value\n");
        return;
    }
    u_int32_t   addr = strtoul(tokens[1], NULL, 0);
    u_int32_t   v = strtoul(tokens[2], NULL, 0);
    
    if (addr > GPMEM_SIZE) {
        printf("address: 0x%x exceeds GPMEM_SIZE\n");
        return;
    }
    ec->data[addr] = v;
}

void _set_breakpoint(FILE *outfp, struct execution_context *ec, char **tokens) {
    // setbp n pc = value
    // setbp n r? = value
    // setbp n addr = value
    if (strlen(tokens[4]) < 1) {
        fprintf(outfp, "usage:\n"
            "\tsetbp n pc = value\n"
            "\tsetbp n r# = value (or <, <=, >, >=) -- for register watch point\n"
            "\tsetbp n # = value (or <, <=, >, >=) -- for memory watch point\n"
            "*Make sure to use space between each symbol in the comparison\n");
        return;
    }
#define SPACE_PC    0
#define SPACE_REG   1
#define SPACE_MEM   2
#define CMP_EQUAL   0
#define CMP_LT      1
#define CMP_LE      2
#define CMP_GT      3
#define CMP_GE      4
    u_int32_t   address;
    int space = SPACE_PC;
    if (!strcmp(tokens[2], "pc")) {
        space = SPACE_PC;  // default
    } else if (tokens[2][0] == 'r' || tokens[1][0] == 'R') {
        space = SPACE_REG;
        address = strtoul(&tokens[2][1], NULL, 0);
        if (address >= NUMBER_OF_REGISTERS) {
            fprintf(outfp, "invalid register: r%d\n", address);
            return;
        }
    } else {
        space = SPACE_MEM;
        address = strtoul(tokens[2], NULL, 0);
        if (address >= GPMEM_SIZE) {
            fprintf(outfp, "memory address: 0x%x > GPMEM_SIZE\n", address);
            return;
        }
    }
    int compare = CMP_EQUAL;
    if (!strcmp(tokens[3], "="))
        compare = CMP_EQUAL;    // default
    else if (!strcmp(tokens[3], "<"))
        compare = CMP_LT;
    else if (!strcmp(tokens[3], "<="))
        compare = CMP_LE;
    else if (!strcmp(tokens[3], ">"))
        compare = CMP_GT;
    else if (!strcmp(tokens[3], ">="))
        compare = CMP_GE;
    else if (!strcmp(tokens[3], "=="))
        compare = 0;
    else {
        fprintf(outfp, "Unknown comparison: %s\n", tokens[3]);
        return;
    }
    
    int i = strtoul(tokens[1], NULL, 0);
    if (i < 0 || i > NUMBER_OF_BREAKPOINTS) {
        fprintf(outfp, "invalid breakpoint: %d\n", i);
        return;
    }
    u_int32_t value = strtoul(tokens[4], NULL, 0);
    
    switch (space) {
        case    SPACE_PC:
            ec->breakpoints[i].breakpoint_type = pc_equal;
            ec->breakpoints[i].value = value;
            break;
        case    SPACE_REG:
            ec->breakpoints[i].address = address;
            ec->breakpoints[i].value = value;
            switch (compare) {
                case    CMP_EQUAL:
                    ec->breakpoints[i].breakpoint_type = reg_value_equal;
                    break;
                case    CMP_LT:
                    ec->breakpoints[i].breakpoint_type = reg_value_lt;
                    break;
                case    CMP_LE:
                    ec->breakpoints[i].breakpoint_type = reg_value_le;
                    break;
                case    CMP_GT:
                    ec->breakpoints[i].breakpoint_type = reg_value_gt;
                    break;
                case    CMP_GE:
                    ec->breakpoints[i].breakpoint_type = reg_value_ge;
                    break;
            }
            break;
        case    SPACE_MEM:
            ec->breakpoints[i].address = address;
            ec->breakpoints[i].value = value;
            switch (compare) {
                case    CMP_EQUAL:
                    ec->breakpoints[i].breakpoint_type = mem_value_equal;
                    break;
                case    CMP_LT:
                    ec->breakpoints[i].breakpoint_type = mem_value_lt;
                    break;
                case    CMP_LE:
                    ec->breakpoints[i].breakpoint_type = mem_value_le;
                    break;
                case    CMP_GT:
                    ec->breakpoints[i].breakpoint_type = mem_value_gt;
                    break;
                case    CMP_GE:
                    ec->breakpoints[i].breakpoint_type = mem_value_ge;
                    break;
            }
            break;
    }
    fprintf(outfp, "breakpoint %d: ", i);
    _print_breakpoint(&ec->breakpoints[i], outfp);
    fprintf(outfp, "\n");
}

void _show_breakpoint(FILE *outfp, struct execution_context *ec, char **tokens) {
    int i;

    if (strlen(tokens[1]) < 1) {
        for (i = 0; i < NUMBER_OF_BREAKPOINTS; i++)
            if (ec->breakpoints[i].breakpoint_type != disabled) {
                fprintf(outfp, "breakpoint %d: ", i);
                _print_breakpoint(&ec->breakpoints[i], outfp);
                fprintf(outfp, "\n");
            }
        return;                
    }
    i = strtoul(tokens[1], NULL, 0);
    if (i < 0 || i >= NUMBER_OF_BREAKPOINTS) {
        fprintf(outfp, "invalid breakpoint %d\n", i);
        return;
    }
    if (ec->breakpoints[i].breakpoint_type != disabled) {
        fprintf(outfp, "breakpoint %d: ", i);
        _print_breakpoint(&ec->breakpoints[i], outfp);
        fprintf(outfp, "\n");
    } else {
        fprintf(outfp, "breakpoint %d: not set\n", i);
    }
}

void _clear_breakpoint(FILE *outfp, struct execution_context *ec, char **tokens) {
    int i;

    if (strlen(tokens[1]) < 1)
        return;                

    i = strtoul(tokens[1], NULL, 0);
    if (i < 0 || i >= NUMBER_OF_BREAKPOINTS) {
        fprintf(outfp, "invalid breakpoint %d\n", i);
        return;
    }
    ec->breakpoints[i].breakpoint_type = disabled;
}

struct debug_command debug_commands[] = {
    { "help", "h", _print_debug_commands, "print basic commands" },
    { "regs", "r", _print_registers, "dump register state" },
    { "quit", "q", _quit, "exit debugger"},
    { "step", "s", _single_step, "single step a single instruction" },
    { "run", "c", _run, "run program to completion" },
    { "setr", "v", _set_register, "set register value" },
    { "disasm", "d", _disasm, "disassemble" },
    { "dump", "m", _dump_memory, "show memory contents" },
    { "setm", "w", _set_memory, "set memory contents" },
    { "setbp", "b", _set_breakpoint, "set breakpoint" },
    { "showbp", "t", _show_breakpoint, "show breakpoints" },
    { "clearbp", "y", _clear_breakpoint, "clear breakpoint" },
    { NULL, NULL, NULL, NULL },
};

int main(int argc, char *argv[]) {
    FILE    *infp = stdin;
    FILE    *outfp = stdout;
    FILE    *debugfp = NULL;
    FILE    *binaryfp = NULL;
    FILE    *sourcefp = NULL;
    FILE    *memoryfp = NULL;
    bool    fallback_to_stdin = false;
    bool    memory_image_is_binary = true;
    int     i;
    struct  execution_context   *ec = malloc(sizeof(*ec));

    memset(ec, 0, sizeof(*ec));
    
    // parse command line
    i = 1;
    while (i < argc) {
        if (argv[i][0] == '-' || argv[i][0] == '/') {
            switch (argv[i][1]) {
                case    'b':
                case    'B':
                    binaryfp = fopen(argv[i + 1], "r");
                    if (!binaryfp) {
                        printf("Failure to open binary image file: %s\n", argv[i + 1]);
                        exit(-__LINE__);
                    }
                    i += 2;
                    break;
                case    'd':
                case    'D':
                    debugfp = fopen(argv[i + 1], "r");
                    if (!debugfp) {
                        printf("Failure to open symbol file: %s\n", argv[i + 1]);
                        exit(-__LINE__);
                    }
                    i += 2;
                    break;
                case    'a':
                case    'A':
                case    'm':
                case    'M':
                    memoryfp = fopen(argv[i + 1], "r");
                    if (!memoryfp) {
                        printf("Failure to open memory image file: %s\n", argv[i + 1]);
                        exit(-__LINE__);
                    }
                    if (argv[i][1] == 'a' || argv[i][1] == 'A')
                        memory_image_is_binary = false;
                    else
                        memory_image_is_binary = true;
                    i += 2;
                    break;
                case    's':
                case    'S':
                    sourcefp = fopen(argv[i + 1], "r");
                    if (!sourcefp) {
                        printf("Failure to open source file: %s\n", argv[i + 1]);
                        exit(-__LINE__);
                    }
                    i += 2;
                    break;
                case    'i':
                case    'I':
                    infp = fopen(argv[i + 1], "r");
                    if (!infp) {
                        printf("Failure to open command input file: %s\n", argv[i + 1]);
                        exit(-__LINE__);
                    }
                    i += 2;
                    break;
                case    'o':
                case    'O':
                    outfp = fopen(argv[i + 1], "w");
                    if (!outfp) {
                        printf("Failure to open output file: %s\n", argv[i + 1]);
                        exit(-__LINE__);
                    }
                    i += 2;
                    break;
                case    'q':
                case    'Q':
                    ec->ctx.queue_src = atoi(argv[i + 1]);
                    i += 2;
                    break;
                case    'c':
                case    'C':
                    fallback_to_stdin = true;
                    ++i;
                    break;
                default:
                    usage_and_exit();
            }
        } else
            usage_and_exit();
    }  
    
    if (sourcefp)
        _read_source_file(ec, sourcefp);

    if (debugfp)
        _read_debug_symbols(ec, debugfp);
    
    if (binaryfp)
        _read_code(ec, binaryfp);
    else
        usage_and_exit();
        
    if (memoryfp) {
        if (memory_image_is_binary)
            _read_memory_binary(ec, memoryfp);
        else
            _read_memory_ascii(ec, memoryfp);
    }

#define MAX_TOKENS  6
    char    old_command_line[MAX_LINE_LENGTH] = "";
    char    debug_command_line[MAX_LINE_LENGTH];
    char    *tokens[MAX_TOKENS];
    for (i = 0; i < MAX_TOKENS; i++)
        tokens[i] = malloc(sizeof(MAX_LINE_LENGTH));
        
    while (!feof(infp)) {
        if (infp == stdin) {
            fprintf(outfp, "db> ");
            fflush(outfp);
        }
        fgets(debug_command_line, MAX_LINE_LENGTH, infp);
        if (feof(infp)) {
            if (fallback_to_stdin) {
                infp = stdin;
                continue;
            } else
                break;
        }
        // If the user just hits enter, then take that as a "redo last command"
        if (strlen(debug_command_line) <= 1)
            strcpy(debug_command_line, old_command_line);
        else
            strcpy(old_command_line, debug_command_line);
        // now parse the command
        _tokenize(debug_command_line, tokens, MAX_TOKENS);
        // now search the debug commands for a command to execute
        i = 0;
        while (debug_commands[i].command) {
            if (!strcmp(tokens[0], debug_commands[i].command) ||
                (strlen(tokens[0]) == 1 &&
                 debug_commands[i].abbreviation[0] == tokens[0][0])) {
                debug_commands[i].func(outfp, ec, tokens);
                break;
            }
            ++i;
        }
        if (!debug_commands[i].command)
            fprintf(outfp, "unknown command: %s -- try \"help\"\n", tokens[0]);
    }
    
    return 0;
}
