#include <stdbool.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include "467gpu.h"
    
struct instruction instructions[] = {
    {   "ldgpmem",  LDGPMEM,    MEMORY | REG3, REG1, 0 },
    {   "ldgpmemi", LDGPMEM,    MEMORY | CONSTANT, REG1, 0 },
    {   "stgpmem",  STGPMEM,    REG1, MEMORY | REG3, 0 },
    {   "stgpmemi", STGPMEM,    REG1, MEMORY | CONSTANT, 0 },
    {   "mul",      MUL,        REG2, REG3, REG1 },
    {   "muli",     MUL,        REG2, CONSTANT, REG1 },
    {   "add",      ADD,        REG2, REG3, REG1 },
    {   "addi",     ADD,        REG2, CONSTANT, REG1 },
    {   "sub",      SUB,        REG2, REG3, REG1 },
    {   "subi",     SUB,        REG2, CONSTANT, REG1 },
    {   "srl",      SRL,        REG2, REG3, REG1 },
    {   "srli",     SRL,        REG2, CONSTANT, REG1 },
    {   "sll",      SLL,        REG2, REG3, REG1 },
    {   "slli",     SLL,        REG2, CONSTANT, REG1 },
    {   "and",      AND,        REG2, REG3, REG1 },
    {   "andi",     AND,        REG2, CONSTANT, REG1 },
    {   "not",      NOT,        REG2, REG1, 0 },
    {   "xor",      XOR,        REG2, REG3, REG1 },
    {   "xori",     XOR,        REG2, CONSTANT, REG1 },
    {   "or",       OR,         REG2, REG3, REG1 },
    {   "ori",      OR,         REG2, CONSTANT, REG1 },
    {   "nand",     NAND,       REG2, REG3, REG1 },
    {   "nandi",    NAND,       REG2, CONSTANT, REG1 },
    {   "li",       LI,         REG1, CONSTANT },
    {   "setlt",    SETLT,      REG2, REG3, REG1 | PREDICATE },
    {   "setlti",   SETLT,      REG2, CONSTANT, REG1 | PREDICATE },
    {   "storeq",   STOREQ,     REG3, 0, 0 },
    {   "storeqi",  STOREQ,     CONSTANT, 0, 0 },
    {   "end",      END,        0, 0, 0 },
    {   NULL,       0,          0, 0, 0 }
};

u_int32_t    encode_instruction(
    u_int32_t   opcode,
    u_int32_t   predicate,
    bool        use_constant,
    u_int32_t   constant,
    u_int32_t   reg1,
    u_int32_t   reg2,
    u_int32_t   reg3) {

    u_int32_t   res = 0;
    
    res = predicate << PREDICATE_SHIFT;
    res |= use_constant << OP3_TYPE_SHIFT;
    res |= opcode << OPCODE_SHIFT;  
    res |= reg1 << REG1_SHIFT;
    res |= reg2 << REG2_SHIFT;
    if (use_constant)
        res |= constant << CONSTANT_SHIFT;
    else
        res |= reg3 << REG3_SHIFT;

    return res;      
}

void decode_instruction(u_int32_t res,
    u_int32_t   *opcode,
    u_int32_t   *predicate,
    bool        *use_constant,
    u_int32_t   *constant,
    u_int32_t   *reg1,
    u_int32_t   *reg2,
    u_int32_t   *reg3) {
    
    *predicate = (res >> PREDICATE_SHIFT) & PREDICATE_MASK;
    *opcode = (res >> OPCODE_SHIFT) & OPCODE_MASK;
    *constant = (res >> CONSTANT_SHIFT) & CONSTANT_MASK;
    *reg1 = (res >> REG1_SHIFT) & REG1_MASK;
    *reg2 = (res >> REG2_SHIFT) & REG2_MASK;
    *reg3 = (res >> REG3_SHIFT) & REG3_MASK;
    *use_constant = ((res >> OP3_TYPE_SHIFT) & OP3_TYPE_MASK) ? true : false;
}

static bool _uses_constant(u_int32_t operand_format) {
    operand_format = operand_format & (~MEMORY) & (~PREDICATE);
    if (operand_format == CONSTANT)
        return true;
    return false;
}

static void _format_operand(char *s, u_int32_t operand_format,
    u_int32_t constant, u_int32_t reg1, u_int32_t reg2, u_int32_t reg3) {
    char  t[128] = "", t1[128];
    
    if (operand_format & MEMORY) {
        sprintf(t, "*");
        operand_format = operand_format & (~MEMORY);
    }
    
    if (operand_format & PREDICATE) {
        operand_format = operand_format & (~PREDICATE);
        switch (operand_format) {
            case    REG1: sprintf(t1, "p%d", reg1); break;
            case    REG2: sprintf(t1, "p%d", reg2); break;
            case    REG3: sprintf(t1, "p%d", reg3); break;
        }
    } else {
        switch (operand_format) {
            case    CONSTANT: sprintf(t1, "0x%4.4x", constant);   break;
            case    REG1: sprintf(t1, "r%d", reg1); break;
            case    REG2: sprintf(t1, "r%d", reg2); break;
            case    REG3: sprintf(t1, "r%d", reg3); break;
        }
    }    
    strcpy(s, t);
    strcat(s, t1);
}

bool disasm_instruction(u_int32_t   res, char *s) {
    u_int32_t   opcode, predicate, constant, reg1, reg2, reg3;
    bool        use_constant;
    char        t1[128], t[128];
    
    decode_instruction(res, &opcode, &predicate, &use_constant,
        &constant, &reg1, &reg2, &reg3);
        
    int ins;
    
    ins = 0;
    while (instructions[ins].mnemonic) {
        if (instructions[ins].opcode == opcode) {
            if (use_constant &&
                (_uses_constant(instructions[ins].operand0) ||
                 _uses_constant(instructions[ins].operand1) ||
                 _uses_constant(instructions[ins].operand2)))
                 break;
            if (!use_constant &&
                !(_uses_constant(instructions[ins].operand0) ||
                 _uses_constant(instructions[ins].operand1) ||
                 _uses_constant(instructions[ins].operand2)))
                break;
        }
        ++ins;
    }
    if (!instructions[ins].mnemonic)
        return false;
    
    sprintf(t1, "%s", instructions[ins].mnemonic);
    if (predicate != 0) {
        sprintf(t, "/%d", predicate);
        strcat(t1, t);
    }
    sprintf(s, "%-10.10s", t1);
    
    bool    first_operand = true;
    if (instructions[ins].operand0) {
        _format_operand(t1, instructions[ins].operand0, constant, reg1, reg2, reg3);
        if (!first_operand)
            strcat(s, ", ");
        strcat(s, t1);
        first_operand = false;
    }
    if (instructions[ins].operand1) {
        _format_operand(t1, instructions[ins].operand1, constant, reg1, reg2, reg3);
        if (!first_operand)
            strcat(s, ", ");
        strcat(s, t1);
        first_operand = false;
    }
    if (instructions[ins].operand2) {
        _format_operand(t1, instructions[ins].operand2, constant, reg1, reg2, reg3);
        if (!first_operand)
            strcat(s, ", ");
        strcat(s, t1);
        first_operand = false;
    }
    return true;
}

u_int32_t   _sign_extend(u_int32_t  constant) {
    if (constant & 0x8000)
        constant = constant | 0xffff0000UL;
    return constant;
}

enum single_step_result_t single_step(struct context *ctx,
    u_int32_t *code, u_int32_t *data,
    u_int32_t   *queue, int *queue_num) {
    u_int32_t   res = code[ctx->pc];
    int         i;
    
    u_int32_t   opcode, predicate, constant, reg1, reg2, reg3;
    bool        use_constant;
    
    decode_instruction(res, &opcode, &predicate, &use_constant,
        &constant, &reg1, &reg2, &reg3);

    // MSR register is read only.  Writes to it are discarded.
    ctx->regs[MSR_REGISTER] = ctx->msr;
        
    // now setup a "virtual" execution environment
    u_int32_t   v2 = 0, v3 = 0, p = 1, wp;
    
    reg1 = reg1 & (~MEMORY) & (~PREDICATE);
    reg2 = reg2 & (~MEMORY) & (~PREDICATE);
    reg3 = reg3 & (~MEMORY) & (~PREDICATE);
    
    v2 = ctx->regs[reg2];
    if (use_constant) {
        switch (opcode) {
            case    LDGPMEM:
            case    STGPMEM:
            case    SRL:
            case    SLL:
            case    AND:
            case    XOR:
            case    OR:
            case    NAND:
            case    STOREQ:
                v3 = constant;
                break;
            default:
                v3 = _sign_extend(constant);
        }
    }
    else
        v3 = ctx->regs[reg3];
    
    if (predicate != 0) {
        switch (predicate) {
            case    1:  p = ctx->predicate1; break;
            case    2:  p = ctx->predicate2; break;
            case    3:  p = ctx->predicate3; break;
        }
    }
     
    if (p) {
        switch (opcode) {
            case    END:
                return at_end;
            case    LDGPMEM:
                if (v3 > GPMEM_SIZE)
                    return invalid_address;
                ctx->regs[reg1] = data[v3];
                break;
            case    STGPMEM:
                if (v3 > GPMEM_SIZE)
                    return invalid_address;
                data[v3] = ctx->regs[reg1];
                break;
            case    MUL:
                ctx->regs[reg1] = v2 * v3;
                break;
            case    ADD:
                ctx->regs[reg1] = v2 + v3;
                break;
            case    SUB:
                ctx->regs[reg1] = v2 - v3;
                break;
            case    SRL:
                ctx->regs[reg1] = v2 >> v3;
                break;
            case    SLL:
                ctx->regs[reg1] = v2 << v3;
                break;
            case    AND:
                ctx->regs[reg1] = v2 & v3;
                break;
            case    NOT:
                ctx->regs[reg1] = ~v2;
                break;
            case    XOR:
                ctx->regs[reg1] = v2 ^ v3;
                break;
            case    OR:
                ctx->regs[reg1] = v2 | v3;
                break;
            case    NAND:
                ctx->regs[reg1] = ~(v2 & v3);
                break;
            case    SETLT:
                if (reg1 >= NUMBER_OF_PREDICATES)
                    return invalid_predicate;
                if (reg1 == 0)
                    break;  // do not write to predicate 0
                if (v2 < v3)
                    wp = 1;
                else
                    wp = 0;
                switch (reg1) {
                    case    1: ctx->predicate1 = wp; break;
                    case    2: ctx->predicate2 = wp; break;
                    case    3: ctx->predicate3 = wp; break;
                }
                break;
            case    STOREQ:
                *queue_num = v3;
                for (i = 0; i < QUEUE_IO; i++)
                    queue[i] = ctx->regs[i];
                ctx->pc++;
                return queue_write;
                break;
            case LI:
                ctx->regs[reg1] = v3;
                break;
            default:
                return -1;
        }
        ctx->pc++;
        return normal_execution;
    } else {
        ctx->pc++;
        return skipped_execution;
    }    
}

